/*
 *  jDTAUS Core RI Client Container
 *  Copyright (c) 2005 Christian Schulte
 *
 *  Christian Schulte, Haldener Strasse 72, 58095 Hagen, Germany
 *  <cs@jdtaus.org> (+49 2331 3543887)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
package org.jdtaus.core.container.ri.client;

import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.WeakHashMap;
import org.jdtaus.core.container.Container;
import org.jdtaus.core.container.ContainerError;
import org.jdtaus.core.container.ContainerInitializer;
import org.jdtaus.core.container.ContextFactory;
import org.jdtaus.core.container.ContextInitializer;
import org.jdtaus.core.container.Dependency;
import org.jdtaus.core.container.DependencyCycleException;
import org.jdtaus.core.container.Implementation;
import org.jdtaus.core.container.MissingImplementationException;
import org.jdtaus.core.container.ModelFactory;
import org.jdtaus.core.container.MultiplicityConstraintException;
import org.jdtaus.core.container.Specification;
import org.jdtaus.core.container.ri.client.versioning.VersionParser;

/**
 * {@code Container} reference implementation.
 *
 * @author <a href="mailto:cs@schulte.it">Christian Schulte</a>
 * @version $Id: DefaultContainer.java 5416 2008-10-27 10:55:51Z schulte2005 $
 *
 * @see org.jdtaus.core.container.ContainerFactory
 */
public class DefaultContainer implements Container
{
    //--Constants---------------------------------------------------------------

    /** Empty Class-Array. */
    private static final Class[] EMPTY =
    {
    };

    /** Implementation Class-Array. */
    private static final Class[] IMPL_CTOR =
    {
        Implementation.class
    };

    /** Dependency Class-Array. */
    private static final Class[] DEP_CTOR =
    {
        Dependency.class
    };

    //---------------------------------------------------------------Constants--
    //--Container---------------------------------------------------------------

    /** Maps implementation identifiers to singleton instances. */
    private final Map singletons = new HashMap( 100 );

    /** Mutex used to detect cyclic instantiations. */
    private final Object cycle = new Object();

    /** Maps objects to model object. */
    private final Map objects = new WeakHashMap( 1024 );

    public Object getObject( final String specificationIdentifier,
        final String implementationName )
    {
        if ( specificationIdentifier == null )
        {
            throw new NullPointerException( "specificationIdentifier" );
        }

        try
        {
            final Object object;
            final Specification specification =
                ModelFactory.getModel().getModules().
                getSpecification( specificationIdentifier );

            final Class specificationClass = this.getClassLoader().
                loadClass( specification.getIdentifier() );

            if ( implementationName != null )
            {
                final Implementation implementation =
                    specification.getImplementation( implementationName );

                final Instance instance =
                    new Instance( specification.getScope(),
                    implementation.getIdentifier(),
                    implementation.getModelVersion() );

                instance.setImplementation( implementation );
                object = this.getObject( instance );
            }
            else
            {
                Instance instance;
                Implementation implementation;
                switch ( specification.getMultiplicity() )
                {
                    case Specification.MULTIPLICITY_ONE:
                        if ( specification.getImplementations().size() != 1 )
                        {
                            throw new MultiplicityConstraintException(
                                specification.getIdentifier() );

                        }

                        implementation = specification.getImplementations().
                            getImplementation( 0 );

                        instance =
                            new Instance( specification.getScope(),
                            implementation.getIdentifier(),
                            implementation.getModelVersion() );

                        instance.setImplementation( implementation );
                        object = this.getObject( instance );
                        break;

                    case Specification.MULTIPLICITY_MANY:
                        final List list =
                            new ArrayList( specification.getImplementations().
                            size() );

                        for ( int i = specification.getImplementations().
                            size() - 1; i >= 0; i-- )
                        {
                            implementation = specification.getImplementations().
                                getImplementation( i );

                            instance =
                                new Instance( specification.getScope(),
                                implementation.getIdentifier(),
                                implementation.getModelVersion() );

                            instance.setImplementation( implementation );
                            list.add( this.getObject( instance ) );
                        }

                        final Object[] implementations =
                            (Object[]) Array.newInstance( specificationClass,
                            list.size() );

                        object = list.toArray( implementations );
                        break;

                    default:
                        throw new AssertionError(
                            Integer.toString(
                            specification.getMultiplicity() ) );

                }
            }

            return object;
        }
        catch ( ClassNotFoundException e )
        {
            throw new ContainerError( e );
        }
    }

    public Object getDependency( final Object object,
        final String dependencyName )
    {
        if ( object == null )
        {
            throw new NullPointerException( "object" );
        }
        if ( dependencyName == null )
        {
            throw new NullPointerException( "dependencyName" );
        }

        try
        {
            final Implementation impl = this.getImplementation( object );
            final Dependency dependency =
                impl.getDependencies().getDependency( dependencyName );

            final Class specificationClass = this.getClassLoader().
                loadClass( dependency.getSpecification().getIdentifier() );

            Instance objectInstance;
            synchronized ( this.objects )
            {
                objectInstance = (Instance) this.objects.get( object );
                if ( objectInstance == null )
                {
                    objectInstance = new Instance( Specification.SCOPE_MULTITON,
                        impl.getIdentifier(),
                        impl.getModelVersion() );

                    objectInstance.setImplementation( impl );
                    this.objects.put( object, objectInstance );
                }
            }

            Object dependencyObject;
            synchronized ( objectInstance )
            {
                dependencyObject =
                    objectInstance.getDependencyObject( dependencyName );

                if ( dependencyObject == null )
                {
                    Dependency clone;
                    Instance dependencyInstance;
                    Implementation implementation;
                    if ( dependency.getImplementation() != null )
                    {
                        dependencyInstance = new Instance(
                            dependency.getSpecification().getScope(),
                            dependency.getImplementation().getIdentifier(),
                            dependency.getImplementation().getModelVersion() );

                        dependencyInstance.setDependency( dependency );

                        dependencyObject = this.getObject( dependencyInstance );

                        if ( dependency.isBound() )
                        {
                            objectInstance.setDependencyObject(
                                dependencyName, dependencyObject );

                        }
                    }
                    else
                    {
                        switch ( dependency.getSpecification().getMultiplicity() )
                        {
                            case Specification.MULTIPLICITY_ONE:
                                if ( dependency.getSpecification().
                                    getImplementations().size() != 1 )
                                {
                                    throw new MultiplicityConstraintException(
                                        dependency.getSpecification().
                                        getIdentifier() );

                                }

                                implementation = dependency.getSpecification().
                                    getImplementations().getImplementation( 0 );

                                dependencyInstance = new Instance(
                                    dependency.getSpecification().getScope(),
                                    implementation.getIdentifier(),
                                    implementation.getModelVersion() );

                                clone = (Dependency) dependency.clone();
                                clone.setImplementation( implementation );
                                dependencyInstance.setDependency( clone );
                                dependencyObject =
                                    this.getObject( dependencyInstance );

                                if ( dependency.isBound() )
                                {
                                    objectInstance.setDependencyObject(
                                        dependencyName, dependencyObject );

                                }
                                break;

                            case Specification.MULTIPLICITY_MANY:
                                final List list =
                                    new ArrayList( dependency.getSpecification().
                                    getImplementations().size() );

                                for ( int i = dependency.getSpecification().
                                    getImplementations().size() - 1; i >= 0; i-- )
                                {
                                    implementation =
                                        dependency.getSpecification().
                                        getImplementations().
                                        getImplementation( i );

                                    dependencyInstance = new Instance(
                                        dependency.getSpecification().
                                        getScope(),
                                        implementation.getIdentifier(),
                                        implementation.getModelVersion() );

                                    clone = (Dependency) dependency.clone();
                                    clone.setImplementation( implementation );
                                    dependencyInstance.setDependency( clone );
                                    list.add(
                                        this.getObject( dependencyInstance ) );

                                }

                                final Object[] implementations =
                                    (Object[]) Array.newInstance(
                                    specificationClass, list.size() );

                                dependencyObject =
                                    list.toArray( implementations );

                                if ( dependency.isBound() )
                                {
                                    objectInstance.setDependencyObject(
                                        dependencyName, dependencyObject );

                                }
                                break;

                            default:
                                throw new AssertionError(
                                    Integer.toString( dependency.
                                    getSpecification().
                                    getMultiplicity() ) );

                        }
                    }
                }
            }

            return dependencyObject;
        }
        catch ( ClassNotFoundException e )
        {
            throw new ContainerError( e );
        }
    }

    public Object getProperty( final Object object,
        final String propertyName )
    {
        if ( object == null )
        {
            throw new NullPointerException( "object" );
        }
        if ( propertyName == null )
        {
            throw new NullPointerException( "propertyName" );
        }

        final Implementation impl = this.getImplementation( object );

        Instance objectInstance;
        synchronized ( this.objects )
        {
            objectInstance = (Instance) this.objects.get( object );
            if ( objectInstance == null )
            {
                objectInstance = new Instance( Specification.SCOPE_MULTITON,
                    impl.getIdentifier(),
                    impl.getModelVersion() );

                objectInstance.setImplementation( impl );
                this.objects.put( object, objectInstance );
            }
        }

        return objectInstance.getProperties().getProperty( propertyName ).
            getValue();

    }

    public String getMessage( final Object object, final String messageName,
        final Object arguments )
    {
        if ( object == null )
        {
            throw new NullPointerException( "object" );
        }
        if ( messageName == null )
        {
            throw new NullPointerException( "messageName" );
        }

        final Implementation impl = this.getImplementation( object );

        Instance objectInstance;
        synchronized ( this.objects )
        {
            objectInstance = (Instance) this.objects.get( object );
            if ( objectInstance == null )
            {
                objectInstance = new Instance( Specification.SCOPE_MULTITON,
                    impl.getIdentifier(),
                    impl.getModelVersion() );

                objectInstance.setImplementation( impl );
                this.objects.put( object, objectInstance );
            }
        }

        final String template = objectInstance.getMessages().
            getMessage( messageName ).getTemplate().
            getValue( Locale.getDefault() );

        final MessageFormat messageFormat =
            new MessageFormat( template, Locale.getDefault() );

        return messageFormat.format( arguments );
    }

    public final Object getImplementation(
        final Class specification, final String implementationName )
    {
        if ( specification == null )
        {
            throw new NullPointerException( "specification" );
        }
        if ( implementationName == null )
        {
            throw new NullPointerException( "implementationName" );
        }

        return this.getObject( specification.getName(), implementationName );
    }

    public final Object getDependency(
        final Class implementation, final String dependencyName )
    {
        if ( implementation == null )
        {
            throw new NullPointerException( "implementation" );
        }
        if ( dependencyName == null )
        {
            throw new NullPointerException( "dependencyName" );
        }

        try
        {
            final Object dependencyObject;
            final Implementation impl =
                ModelFactory.getModel().getModules().
                getImplementation( implementation.getName() );

            final Dependency dependency =
                impl.getDependencies().getDependency( dependencyName );

            final Class specificationClass = this.getClassLoader().
                loadClass( dependency.getSpecification().getIdentifier() );

            if ( dependency.getImplementation() != null )
            {
                final Instance dependencyInstance =
                    new Instance( dependency.getSpecification().getScope(),
                    dependency.getImplementation().
                    getIdentifier(),
                    dependency.getImplementation().
                    getModelVersion() );

                dependencyInstance.setDependency( dependency );
                dependencyObject = this.getObject( dependencyInstance );
            }
            else
            {
                Dependency clone;
                Implementation dependencyImpl;
                Instance dependencyInstance;
                switch ( dependency.getSpecification().getMultiplicity() )
                {
                    case Specification.MULTIPLICITY_ONE:
                        if ( dependency.getSpecification().
                            getImplementations().size() != 1 )
                        {
                            throw new MultiplicityConstraintException(
                                dependency.getSpecification().
                                getIdentifier() );

                        }

                        dependencyImpl = dependency.getSpecification().
                            getImplementations().getImplementation( 0 );

                        dependencyInstance = new Instance(
                            dependency.getSpecification().getScope(),
                            dependencyImpl.getIdentifier(),
                            dependencyImpl.getModelVersion() );

                        clone = (Dependency) dependency.clone();
                        clone.setImplementation( dependencyImpl );
                        dependencyInstance.setDependency( clone );
                        dependencyObject = this.getObject( dependencyInstance );
                        break;

                    case Specification.MULTIPLICITY_MANY:
                        final List list =
                            new ArrayList( dependency.getSpecification().
                            getImplementations().size() );

                        for ( int i = dependency.getSpecification().
                            getImplementations().size() - 1; i >= 0; i-- )
                        {
                            dependencyImpl = dependency.getSpecification().
                                getImplementations().getImplementation( i );

                            dependencyInstance = new Instance(
                                dependency.getSpecification().getScope(),
                                dependencyImpl.getIdentifier(),
                                dependencyImpl.getModelVersion() );

                            clone = (Dependency) dependency.clone();
                            clone.setImplementation( dependencyImpl );
                            dependencyInstance.setDependency( clone );
                            list.add( this.getObject( dependencyInstance ) );
                        }

                        final Object[] implementations =
                            (Object[]) Array.newInstance(
                            specificationClass, list.size() );

                        dependencyObject = list.toArray( implementations );
                        break;

                    default:
                        throw new AssertionError(
                            Integer.toString( dependency.getSpecification().
                            getMultiplicity() ) );

                }
            }

            this.initializeContext( dependencyObject );
            return dependencyObject;
        }
        catch ( ClassNotFoundException e )
        {
            throw new ContainerError( e );
        }
    }

    //---------------------------------------------------------------Container--
    //--DefaultContainer--------------------------------------------------------

    /**
     * Gets the classloader used for loading classes.
     * <p>This method returns the current thread's context classloader,
     * if available, or this classes classloader, if the current thread has no
     * context classloader set.</p>
     *
     * @return the classloader to be used for loading classes.
     */
    protected ClassLoader getClassLoader()
    {
        ClassLoader classLoader = Thread.currentThread().
            getContextClassLoader();

        if ( classLoader == null )
        {
            classLoader = this.getClass().getClassLoader();
        }

        return classLoader;
    }

    /**
     * Creates a new instance of an implementation.
     * <p>The reference implementation will load the class corresponding to
     * the implementation by using the classloader returned by method
     * {@link #getClassLoader()} and will then create a new instance of that
     * class by calling the constructor taking an {@code Implementation} as
     * a parameter.</p>
     *
     * @param impl the implementation to return a new instance for.
     *
     * @return a new instance of the class corresponding to {@code impl}.
     *
     * @throws NullPointerException if {@code impl} is {@code null}.
     * @throws org.jdtaus.core.container.InstantiationException
     * if no instance can be created.
     * @deprecated Replaced by {@link #instantiateObject(org.jdtaus.core.container.ri.client.Instance)
     */
    private Object instantiateImplementation( final Implementation impl )
    {
        if ( impl == null )
        {
            throw new NullPointerException( "impl" );
        }

        try
        {
            final Class clazz =
                this.getClassLoader().loadClass( impl.getIdentifier() );

            final Constructor ctor = clazz.getDeclaredConstructor( IMPL_CTOR );
            ctor.setAccessible( true );

            final Object object = ctor.newInstance( new Object[]
                {
                    impl
                } );

            return object;
        }
        catch ( Throwable e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                impl.getIdentifier(), e );

        }
    }

    /**
     * Creates a new instance of a dependency.
     * <p>The reference implementation will load the class corresponding to
     * the dependency by using the classloader returned by method
     * {@link #getClassLoader()} and will then create a new instance of that
     * class by calling the constructor taking a {@code Dependency} as
     * a parameter.</p>
     *
     * @param dep the dependency to return a new instance for.
     *
     * @return a new instance of the class corresponding to {@code dep}.
     *
     * @throws NullPointerException if {@code dep} is {@code null}.
     * @throws org.jdtaus.core.container.InstantiationException
     * if no instance can be created.
     * @deprecated Replaced by {@link #instantiateObject(org.jdtaus.core.container.ri.client.Instance)
     */
    private Object instantiateDependency( final Dependency dep )
    {
        if ( dep == null )
        {
            throw new NullPointerException( "dep" );
        }

        try
        {
            final Class clazz = this.getClassLoader().loadClass(
                dep.getImplementation().getIdentifier() );

            final Constructor ctor =
                clazz.getDeclaredConstructor( DEP_CTOR );

            ctor.setAccessible( true );

            final Object object = ctor.newInstance( new Object[]
                {
                    dep
                } );

            return object;
        }
        catch ( Throwable e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                dep.getImplementation().getIdentifier(), e );

        }
    }

    /**
     * Performs initialization of an object.
     *
     * @param object the object to initialize.
     *
     * @throws NullPointerException if {@code object} is {@code null}.
     * @throws Exception if initialization fails.
     */
    private void initializeObject( final Object object ) throws Exception
    {
        if ( object == null )
        {
            throw new NullPointerException( "object" );
        }

        if ( object instanceof ContainerInitializer )
        {
            ( (ContainerInitializer) object ).initialize();
        }
    }

    /**
     * Performs context initialization of an object.
     *
     * @param object the object to perform context initialization with.
     *
     * @throws NullPointerException if {@code object} is {@code null}.
     * @throws org.jdtaus.core.container.ContextError for unrecoverable context
     * errors.
     */
    private void initializeContext( final Object object )
    {
        if ( object == null )
        {
            throw new NullPointerException( "object" );
        }

        if ( object instanceof ContextInitializer &&
            !( (ContextInitializer) object ).isInitialized(
            ContextFactory.getContext() ) )
        {
            ( (ContextInitializer) object ).initialize(
                ContextFactory.getContext() );

        }
    }

    private Object getObject( final Instance instance )
    {
        Object object = null;
        Throwable throwable = null;

        switch ( instance.getScope() )
        {
            case Specification.SCOPE_MULTITON:
                // Create a new instance.
                object = this.instantiateObject( instance );
                break;

            case Specification.SCOPE_CONTEXT:
                // Return the context instance for impl.
                object = ContextFactory.getContext().
                    getObject( instance.getClassName() );

                if ( object == null || object instanceof Instance )
                {
                    synchronized ( this.cycle )
                    {
                        object = ContextFactory.getContext().
                            getObject( instance.getClassName() );

                        if ( object == null )
                        {
                            ContextFactory.getContext().
                                setObject( instance.getClassName(), instance );

                            try
                            {
                                object = this.instantiateObject( instance );
                            }
                            catch ( Throwable t )
                            {
                                throwable = t;
                            }

                            if ( throwable != null )
                            {
                                ContextFactory.getContext().removeObject(
                                    instance.getClassName() );

                                if ( throwable instanceof Error )
                                {
                                    throw (Error) throwable;
                                }
                                else if ( throwable instanceof RuntimeException )
                                {
                                    throw (RuntimeException) throwable;
                                }
                                else
                                {
                                    throw new org.jdtaus.core.container.InstantiationException(
                                        instance.getClassName(), throwable );

                                }
                            }

                            ContextFactory.getContext().
                                setObject( instance.getClassName(), object );

                        }
                        else if ( object instanceof Instance )
                        {
                            throw new DependencyCycleException(
                                ( (Instance) object ).getClassName(),
                                instance.getClassName() );

                        }
                    }
                }
                break;

            case Specification.SCOPE_SINGLETON:
                // Return the singleton instance for impl.
                object = this.singletons.get( instance.getClassName() );

                if ( object == null || object instanceof Instance )
                {
                    synchronized ( this.cycle )
                    {
                        object = this.singletons.get( instance.getClassName() );

                        if ( object == null )
                        {
                            this.singletons.put( instance.getClassName(),
                                instance );

                            try
                            {
                                object = this.instantiateObject( instance );
                            }
                            catch ( Throwable t )
                            {
                                throwable = t;
                            }

                            if ( throwable != null )
                            {
                                this.singletons.remove( instance.getClassName() );

                                if ( throwable instanceof Error )
                                {
                                    throw (Error) throwable;
                                }
                                else if ( throwable instanceof RuntimeException )
                                {
                                    throw (RuntimeException) throwable;
                                }
                                else
                                {
                                    throw new org.jdtaus.core.container.InstantiationException(
                                        instance.getClassName(), throwable );

                                }
                            }

                            this.singletons.put( instance.getClassName(),
                                object );

                        }
                        else if ( object instanceof Instance )
                        {
                            throw new DependencyCycleException(
                                ( (Instance) object ).getClassName(),
                                instance.getClassName() );

                        }
                    }
                }
                break;

            default:
                throw new AssertionError(
                    Integer.toString( instance.getScope() ) );

        }

        return object;
    }

    private Object instantiateObject( final Instance instance )
    {
        try
        {
            final Class clazz =
                this.getClassLoader().loadClass( instance.getClassName() );

            Object object = null;

            if ( VersionParser.compare( instance.getModelVersion(),
                "1.3" ) < 0 )
            {
                if ( instance.getImplementation() != null )
                {
                    object = this.instantiateImplementation(
                        instance.getImplementation() );

                }
                else if ( instance.getDependency() != null )
                {
                    object = this.instantiateDependency(
                        instance.getDependency() );

                }
                else
                {
                    throw new AssertionError();
                }
            }
            else
            {
                clazz.getConstructor( EMPTY );
                object = clazz.newInstance();
            }

            synchronized ( this.objects )
            {
                this.objects.put( object, instance );
            }

            this.initializeObject( object );
            return object;
        }
        catch ( Throwable t )
        {
            if ( t instanceof Error )
            {
                throw (Error) t;
            }
            else if ( t instanceof RuntimeException )
            {
                throw (RuntimeException) t;
            }
            else
            {
                throw new org.jdtaus.core.container.InstantiationException(
                    instance.getClassName(), t );

            }
        }
    }

    private Implementation getImplementation( final Object object )
    {
        MissingImplementationException exception = null;
        Implementation implementation = null;
        Class clazz = object.getClass();

        do
        {
            try
            {
                implementation = ModelFactory.getModel().getModules().
                    getImplementation( clazz.getName() );

                break;
            }
            catch ( MissingImplementationException e )
            {
                if ( exception == null )
                {
                    exception = e;
                }
            }
        }
        while ( ( clazz = clazz.getSuperclass() ) != null );

        if ( implementation == null )
        {
            throw exception;
        }

        return implementation;
    }

    //--------------------------------------------------------DefaultContainer--
}
